import MPI.*;

public class wavy
{

static final int HOST =0;
static final int MAXPROC=8;
static final int MAXPOINTS = 1000;
static final int MAXSTEPS=10000;
static final double PI = 3.14159265;
static final int E_RtoL = 10;
static final int E_LtoR = 20;
static final int E_OUT1 = 30;
static final int E_OUT2 = 40;

static int processID;         /* task ID */
static int nproc;             /* number of processes */
static int nsteps,            /* number of time steps */
    tpoints,                  /* total points along string*/
    npoints,                  /* number of points handled by
                                 this processor */
    first,                    /* index of 1st point handled
                                 by this processor */
   rcode;                     /* generic return code */
static double values[]=new double[MAXPOINTS+2],  /* values at time t */
              oldval[]=new double[MAXPOINTS+2],  /* values at time (t-dt) */
              newval[]=new double[MAXPOINTS+2];  /* values at time (t+dt) */

 /*--------------------------------------------------------
 *      Master obtains input values from user
 *
 --------------------------------------------------------*/

/*
---------------------------------------------------------
 *      Main program
 *
 --------------------------------------------------------*/

public static void main(String[] args)
  {
    int left, right;
    Status stat;
    String host;
     /* learn number of tasks and process ID */
    MPI.Init(args);
    host= MPI.Get_processor_name();   
    nproc=MPI.COMM_WORLD.Size();
    processID=MPI.COMM_WORLD.Rank();
     
    
     /* Identify the  left and right processes */
    if (processID == nproc-1) right =0;
    else right = processID + 1;
    if (processID == 0) left = nproc- 1;
    else left = processID - 1;
    System.out.println("PID="+processID+"right="+right+"left="+left);
   
     /* get program parameters and initialize wave values */   
    
    if (processID == 0) init_host();
     
    else init_slaves();
    init_wave();

     /* update values along the line for nstep time steps */
     
    update(left, right);

     /* collect results and print */
    if (processID == 0) output_host();
    else output_slaves();

    MPI.Finalize();
  }
  

public static void init_host()
     {
     int i,j;
     int buffer_host[]=new int [2];

     /* set number of points, number of iterations */
     tpoints = 40;
     nsteps = 3;
     

     System.out.println ("PID="+processID+ "points ="+tpoints+" steps ="
     +nsteps);

     /* then broadcast time advance parameter, total
     points, time steps */
     buffer_host[0] = tpoints;
     buffer_host[1] = nsteps;
     for (j=1;j<=nproc;j++)
    
     MPI.COMM_WORLD.Send(buffer_host,0,2,MPI.INT,j, HOST);
     }

/*
-----------------------------------------------------------
--------------
 *      Workers receive input values from master
 *
 ----------------------------------------------------------
 ---------------*/
  
public static  void init_slaves()
     {
     int buffer_slav[]=new int[2];
     int i;
     /* then receive total points, time steps */
     
     
     MPI.COMM_WORLD.Recv(buffer_slav,0,2,MPI.INT,HOST,HOST);
          
     tpoints = buffer_slav[0];
     nsteps = buffer_slav[1];
     System.out.println("PID="+processID+"points="+tpoints+"steps="+nsteps);
     }

/*
-----------------------------------------------------------
-------------
 *     All processes initialize points on line
 *
 ----------------------------------------------------------
 ----------- */

public static void init_wave()
     {
     int  i, j, k;
     double x, fac;

     /* calculate initial values based on sine curve */
     npoints = tpoints/nproc;
     fac = 2.0 * PI;
     for (i = 0, k = 0; i < nproc; i++)
          {
          
          if (processID == i)
               {
               first = k + 1;
               
               for (j = 1; j <= npoints; j++, k++)
                   {
                   x = (double)k/(double)(tpoints - 1);
                   values[j] = Math.sin (fac * x);
                   }  
              

               }
          else k += npoints;
          }
      
     for (i = 1; i <= npoints; i++)
      { oldval[i] = values[i];
      }
     }

/*
-----------------------------------------------------------
--------------
 *      Calculate new values using wave equation
 *
 ----------------------------------------------------------
 ---------------*/
  
public static void do_math(int i)
     {
     final double dtime = 0.3;
     final double c = 1.0;
     final double dx = 1.0; 
     double tau, sqtau;

     tau = (c * dtime / dx);
     sqtau = tau * tau;
     newval[i] = (2.0 * values[i]) - oldval[i]  
                + (sqtau * (values[i-1] - (2.0 * values[i])
                + values[i+1]));
     }

/*
------------------------------------------------------
 *      All processes update their points a specified
 number of times  
 *
 ---------------------------------------------------*/
  
public static void update(int left, int right)
     {
     int i, j; 
     double buffer_h1[]=new double[3];
     double buffer_h2[]=new double[3];
     double buffer_s1[]=new double[3];
     double buffer_s2[]=new double[3];

     for (i = 1; i <= nsteps; i++)
          {
          /* exchange data with "left-hand" neighbor */
          if (processID != 0)  
               {
               
               buffer_h1[0]=values[1];
               MPI.COMM_WORLD.Send(buffer_h1,0,1,MPI.DOUBLE,left,E_RtoL);
               
               
               MPI.COMM_WORLD.Recv(buffer_h2,0,1,MPI.DOUBLE,left,E_LtoR);
                              
               
               values[0]=buffer_h2[0];

               
               }
          /* exchange data with "right-hand" neighbor */
          if (processID != nproc-1)
               {
               
             
               
               MPI.COMM_WORLD.Recv(buffer_s1,0,1,MPI.DOUBLE,right,E_RtoL);
               values[npoints+1]=buffer_s1[0];
               
             
               buffer_s2[0]=values[npoints];
               
               MPI.COMM_WORLD.Send(buffer_s2,0,1,MPI.DOUBLE,right,E_LtoR);
                  
               
                             
               }
                
        
        
          for (j = 1; j <= npoints; j++)
               {
               if (first + j - 1 == 1) 
                 newval[j] = 0.0;
               if (first + j - 1 == tpoints) 
                    newval[j] = 0.0;
               else
                    do_math(j);
               
               }
          
          for (j = 1; j <= npoints; j++)
               
               {
               oldval[j] = values[j];
               values[j] = newval[j];
               }
          }
     }

/*
-----------------------------------------------------------
-------------
 *      Master receives results from workers and prints
 *
 ----------------------------------------------------------
 -------------- */

public static void output_host()  
     {
     int i,j, start,tpts;
     int buffer_h3[]=new int[3];
     double buffer_h4[]=new double [npoints];
     
     double results[]=new double [MAXPOINTS];

     /* store worker's results in results array */
     for (i = 1; i < nproc; i++)
     
       {
          /* receive first point */
          
          MPI.COMM_WORLD.Recv( buffer_h3,0,1,MPI.DOUBLE,i,E_OUT1);
          
          start = buffer_h3[0];
          
          /* receive results */
          MPI.COMM_WORLD.Recv(buffer_h4,0,npoints,MPI.DOUBLE,i,E_OUT2);
          for (j=1;j<=npoints;j++)
          {
           results[start+j-1]=buffer_h4[j];
          }           
       }

     /* store master's results in results array */
     for (i = first; i < first + npoints; i++)
          results[i] = values[i];

     tpts = (tpoints < 25) ? tpoints: 25;
     System.out.println("first"+tpts+"points=");
     
     
     for (i = 0; i < tpts; i++)
     System.out.println(results[i]+i);
      
     
     }

/*
-----------------------------------------------------------
--------------
 *      Slaves send the updated values to host
 *
 ----------------------------------------------------------
 ---------------*/
  
public static void output_slaves()
     {
     int buffer_s3[]=new int [1];
     int j;
     double buffer_s4[]=new double [npoints];
     
     /* send first point host */
     buffer_s3[0] = first;
     MPI.COMM_WORLD.Send(buffer_s3,0,1,MPI.INT, HOST, E_OUT1);
      
     /* send results to host */
     for (j=1;j<=npoints;j++)
      {
      buffer_s4[j]=values[j];
      }
      MPI.COMM_WORLD.Send(buffer_s4,0,npoints,MPI.DOUBLE, HOST,E_OUT2);
      
     }

}

